@@ -1,10 +1,8 @@
-import biplist # binary only
-import plistlib # plain text only
-from importlib import import_module
 import struct
 import os
 import sys
 import textwrap
+from importlib import import_module
 import pprint
 import tempfile
 import sqlite3
@@ -15,13 +13,17 @@ import logging
 from datetime import datetime, timezone
 from pathlib import Path
 
+# import biplist # binary only
+import plistlib # plain text only
+import NSKeyedUnArchiver
+
 try:
     from Cryptodome.Cipher import AES
 except:
     from Crypto.Cipher import AES # https://www.dlitz.net/software/pycrypto/
 
 
-__version__ = '0.9.912'
+__version__ = '0.9.921'
 
 module_logger = logging.getLogger(__name__)
 
@@ -135,7 +137,8 @@ class iOSbackup(object):
 
     def close(self):
         try:
-            os.remove(self.manifestDB)
+            if self.manifestDB is not None:
+                os.remove(self.manifestDB)
         except FileNotFoundError:
             # Its OK if manifest temporary file is not there anymore
             pass
@@ -225,7 +228,7 @@ class iOSbackup(object):
             classKeys=pprint.pformat(self.classKeys, indent=4),
             IsEncrypted=self.manifest['IsEncrypted'],
             PasscodeSet=self.manifest['WasPasscodeSet'],
-            ManifestKey=self.manifest['ManifestKey'].hex(),
+            ManifestKey='Not applicable' if iOSbackup.isOlderThaniOS10dot2(self.manifest['Lockdown']['ProductVersion']) else self.manifest['ManifestKey'].hex(),
             Applications=pprint.pformat(self.manifest['Applications'], indent=4),
             manifestDB=self.manifestDB,
             name=self.manifest['Lockdown']['DeviceName'],
@@ -297,7 +300,8 @@ class iOSbackup(object):
             info={}
             try:
                 with open(manifestFile, 'rb') as infile:
-                    manifest = biplist.readPlist(infile)
+#                     manifest = biplist.readPlist(infile)
+                    manifest = plistlib.load(infile)
             except FileNotFoundError:
                 logging.warning(f"{udid} under {root} doesn't seem to have a manifest file.")
                 return None
@@ -652,7 +656,8 @@ class iOSbackup(object):
         fileList=[]
         for payload in backupFiles:
             payload=dict(payload)
-            payload['manifest']=biplist.readPlistFromString(payload['file'])
+#             payload['manifest']=biplist.readPlistFromString(payload['file'])
+            payload['manifest']=plistlib.loads(payload['file'])
             del payload['file']
 
             # Compute target file with path
@@ -737,7 +742,9 @@ class iOSbackup(object):
 
         if backupFile:
             payload=dict(backupFile)
-            payload['manifest']=biplist.readPlistFromString(payload['file'])
+#             payload['manifest']=biplist.readPlistFromString(payload['file'])
+#             payload['manifest']=plistlib.loads(payload['file'])
+            payload['manifest']=NSKeyedUnArchiver.unserializeNSKeyedArchiver(payload['file'])
             del payload['file']
         else:
             if relativePath:
@@ -760,9 +767,13 @@ class iOSbackup(object):
             manifest=manifestData
         elif type(manifestData)==bytes:
             # Interpret data stream and convert into a dict
-            manifest = biplist.readPlistFromString(manifestData)
+#             manifest = biplist.readPlistFromString(manifestData)
+#             manifest = plistlib.loads(manifestData)
+            manifest = NSKeyedUnArchiver.unserializeNSKeyedArchiver(manifestData)
 
-        fileData=manifest['$objects'][manifest['$top']['root'].integer]
+
+
+        fileData=manifest
 
         folder=True
         if 'EncryptionKey' in fileData:
@@ -793,11 +804,13 @@ class iOSbackup(object):
 
         info=iOSbackup.getFileInfo(manifestData)
 
-        fileData=info['completeManifest']['$objects'][info['completeManifest']['$top']['root'].integer]
+#         fileData=info['completeManifest']['$objects'][info['completeManifest']['$top']['root'].integer]
+        fileData=info['completeManifest']
 
         dataDecrypted=None
         if 'EncryptionKey' in fileData:
-            encryptionKey=info['completeManifest']['$objects'][fileData['EncryptionKey'].integer]['NS.data'][4:]
+#             encryptionKey=info['completeManifest']['$objects'][fileData['EncryptionKey'].integer]['NS.data'][4:]
+            encryptionKey=info['completeManifest']['EncryptionKey'][4:]
 
 
             # {BACKUP_ROOT}/{UDID}/ae/ae2c3d4e5f6...
@@ -936,7 +949,8 @@ class iOSbackup(object):
             return None
 
 
-        fileData=info['completeManifest']['$objects'][info['completeManifest']['$top']['root'].integer]
+#         fileData=info['completeManifest']['$objects'][info['completeManifest']['$top']['root'].integer]
+        fileData=info['completeManifest']
 
 
         if targetName:
@@ -955,7 +969,7 @@ class iOSbackup(object):
 
 
         if 'EncryptionKey' in fileData:
-            encryptionKey=info['completeManifest']['$objects'][fileData['EncryptionKey'].integer]['NS.data'][4:]
+            encryptionKey=fileData['EncryptionKey'][4:]
             key = self.unwrapKeyForClass(fileData['ProtectionClass'], encryptionKey)
 
             chunkSize=16*1000000 # 16MB chunk size
@@ -965,7 +979,10 @@ class iOSbackup(object):
 
             # {BACKUP_ROOT}/{UDID}/ae/ae2c3d4e5f6...
             with open(os.path.join(self.backupRoot, self.udid, fileNameHash[:2], fileNameHash), 'rb') as inFile:
-                mappedInFile = mmap.mmap(inFile.fileno(), length=0, prot=mmap.PROT_READ)
+                if os.name == 'nt':
+                    mappedInFile = mmap.mmap(inFile.fileno(), length=0, access=mmap.ACCESS_READ)
+                else:
+                    mappedInFile = mmap.mmap(inFile.fileno(), length=0, prot=mmap.PROT_READ)
 
                 with open(targetFileName, 'wb') as outFile:
 
@@ -1031,12 +1048,17 @@ class iOSbackup(object):
         with open(os.path.join(self.backupRoot,self.udid,iOSbackup.catalog['manifestDB']), 'rb') as db:
             encrypted_db = db.read()
 
-        manifest_class = struct.unpack('<l', self.manifest['ManifestKey'][:4])[0]
-        manifest_key   = self.manifest['ManifestKey'][4:]
+        # Before iOS 10.2 the manifest database was not encrypted in the backups. So, there is no need for decryption.
+        # Also, the ManifestKey is not presented in the Plist, so all references to 'ManifestKey' would result in KeyError
+        if iOSbackup.isOlderThaniOS10dot2(self.manifest['Lockdown']['ProductVersion']):
+            decrypted_data = encrypted_db
+        else:
+            manifest_class = struct.unpack('<l', self.manifest['ManifestKey'][:4])[0]
+            manifest_key   = self.manifest['ManifestKey'][4:]
 
-        key = self.unwrapKeyForClass(manifest_class, manifest_key)
+            key = self.unwrapKeyForClass(manifest_class, manifest_key)
 
-        decrypted_data = iOSbackup.AESdecryptCBC(encrypted_db, key)
+            decrypted_data = iOSbackup.AESdecryptCBC(encrypted_db, key)
 
 #         print(len(decrypted_data))
 
@@ -1055,7 +1077,8 @@ class iOSbackup(object):
         self.date=iOSbackup.convertTime(os.path.getmtime(manifestFile), since2001=False)
 
         with open(manifestFile, 'rb') as infile:
-            self.manifest = biplist.readPlist(infile)
+#             self.manifest = biplist.readPlist(infile)
+            self.manifest = plistlib.load(infile)
 
 
 
@@ -1069,7 +1092,8 @@ class iOSbackup(object):
 
         statusFile = os.path.join(self.backupRoot, self.udid, iOSbackup.catalog['status'])
         with open(statusFile, 'rb') as infile:
-            self.status = biplist.readPlist(infile)
+#             self.status = biplist.readPlist(infile)
+            self.status = plistlib.load(infile)
 
 
 
@@ -1112,10 +1136,13 @@ class iOSbackup(object):
         except:
             hlib = import_module('hashlib')
 
-
-        temp = hlib.pbkdf2_hmac('sha256', cleanpassword,
-            self.attrs[b"DPSL"],
-            self.attrs[b"DPIC"], 32)
+        # if the ios is older that 10.2, the stage with the DPSL and DPIC are not used in the key derivation
+        if iOSbackup.isOlderThaniOS10dot2(self.manifest['Lockdown']['ProductVersion']):
+            temp = cleanpassword
+        else:
+            temp = hlib.pbkdf2_hmac('sha256', cleanpassword,
+                self.attrs[b"DPSL"],
+                self.attrs[b"DPIC"], 32)
 
         self.decryptionKey = hlib.pbkdf2_hmac('sha1', temp,
             self.attrs[b"SALT"],
@@ -1133,8 +1160,10 @@ class iOSbackup(object):
 
             if classkey[b"WRAP"] & self.WRAP_PASSCODE:
                 k = iOSbackup.AESUnwrap(self.decryptionKey,classkey[b"WPKY"])
+
                 if not k:
-                    return False
+                    raise Exception('Failed decrypting backup. Try to start over with a clear text decrypting password on parameter "cleartextpassword".')
+
                 classkey[b"KEY"] = k
 
         return True
@@ -1146,8 +1175,8 @@ class iOSbackup(object):
         if len(persistent_key) != 0x28:
             raise Exception("Invalid key length")
 
-#         print(f"class: {protection_class}, key: {persistent_key}")
         ck = self.classKeys[protection_class][b"KEY"]
+
         return iOSbackup.AESUnwrap(ck, persistent_key)
 
 
@@ -1233,3 +1262,27 @@ class iOSbackup(object):
             yield (tag,data)
             i += 8 + length
 
+
+
+
+    def isOlderThaniOS10dot2(version):
+        """Return boolean whether the version is older than iOS 10.2
+
+        Parameters
+        ----------
+        version : str,
+            Version we want to compare. Assumes version is separated using point.
+        """
+
+        versions = version.split('.')
+        if int(versions[0])<10:
+            return True
+        if int(versions[0])>10:
+            return False
+        if int(versions[0])==10:
+            if len(versions)==1: #str is ios 10 only
+                return True
+            if int(versions[1])<2:
+                return True
+            else:
+                return False
\ No newline at end of file
